/*
 * linux/arch/mips/vr41xx/ite8181.c
 *
 * Platform support for the machine which LCD controler is ITE 8181
 *
 * Copyright (C) 2000 kei
 * Copyright (C) 2000 SATO Kazumi
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/kernel.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/vr41xx.h>
#include <asm/vr41xx-platdep.h>

/*
#define DEBUG
*/

/* ITE8181 configuration registers */
#define ITE8181_CONF(index) (&ite8181_config_base[(index)/sizeof(unsigned int)])

#define ITE8181_ID       0x00	// Device ID, Vender ID
#define		ITE8181_DATA_ID 0x81811283
#define ITE8181_SCMD     0x04	// Status, Command Reg.
#define ITE8181_CLASS    0x08	// Class, Sub-class, PRG, revision
#define		ITE8181_DATA_CLASS      0x03800000
#define ITE8181_MBA      0x10	// Memory Base Address(4MB bounday)
#define ITE8181_GBA      0x14	// GUI Base Adress(32KB boundary)
#define ITE8181_SBA      0x18	// Graphic Base Address (64KB boundary)
#define ITE8181_TEST     0x40	// Test Reg.
#define ITE8181_STANDBY  0x44	// Standby Reg.
#define 	ITE8181_DATA_PLL2_TEST		0xc0000	// PLL2 is test mode
#define 	ITE8181_DATA_PLL1_TEST		0x30000	// PLL1 is test mode
#define 	ITE8181_DATA_PLL2_RESET		0x8000	// PLL2 reset
#define 	ITE8181_DATA_PLL1_RESET		0x4000	// PLL1 reset
#define 	ITE8181_DATA_PLL2_PWDOWN	0x2000	// PLL2 powerdown
#define 	ITE8181_DATA_PLL1_PWDOWN	0x1000	// PLL1 powerdown
#define		ITE8181_DATA_PALETTESTBY	0x0200	// Pallete RAM standby
#define		ITE8181_DATA_CURSORSTBY		0x0100	// Cursor standby
#define		ITE8181_DATA_BITBLTSTBY		0x0080	// BitBlt engine standby
#define		ITE8181_DATA_LINESTBY		0x0040	// Line Draw standby
#define		ITE8181_DATA_DACCLKSTOP		0x0020	// DAC Clock stop
#define		ITE8181_DATA_DACPOWERON		0x0010	// DAC Power ON
#define		ITE8181_DATA_GATEPLL2IN		0x0008	// Gate PLL2 input clock
#define		ITE8181_DATA_GATEPLL1IN		0x0004	// Gate PLL1 input clock
#define		ITE8181_DATA_CLOCKSTOP		0x0001	// 14.318MHZ CLock Stop
#define ITE8181_PLL1     0x48	// PLL1 Reg.
#define ITE8181_PLL2     0x4c	// PLL2 Reg.

/* ITE8181 GUI 32bit registers */
#define ITE8181_GUI32(index)	(volatile int *)(&ite8181_gui_base[(index)])

#define ITE8181_GUI_BSLE	0x00	// BitBlt src/Line Draw End
#define ITE8181_GUI_BDLS	0x04	// BitBlt dst/Line Draw Start
#define ITE8181_GUI_BPOA	0x08	// BitBlt Pattern Offset Address
#define ITE8181_GUI_BWH		0x0c	// BitBlt Width, Hight
#define ITE8181_GUI_BSO		0x10	// BitBlt Screen Offset
#define ITE8181_GUI_FCR		0x14	// ForeGround Color Reg.
#define ITE8181_GUI_BCR		0x18	// BackGround Color Reg.
#define ITE8181_GUI_BC		0x1c	// BitBlt Control
#define ITE8181_GUI_BS		0x20	// BitBlt Status
#define ITE8181_GUI_ASDS	0x24	// Line Draw Axial Step, Diagonal Step
#define ITE8181_GUI_LET		0x28	// Line Draw Error Term/ Pixel Count
#define ITE8181_GUI_LST		0x2c	// Scissor Top
#define ITE8181_GUI_LSB		0x30	// Scissor Bottom
#define ITE8181_GUI_LSR		0x34	// Line Style Register
#define ITE8181_GUI_SSVS	0x38	// Short Stroke Vector Spec
#define ITE8181_GUI_MR		0x4c	// Misc Reg.
#define ITE8181_GUI_PIO		0x40000	// Pixel I/port for System Data

/* ITE8181 GUI 8bit registers */	
#define ITE8181_GUI8(index)	(&ite8181_gui_base[(index)])

#define	ITE8181_GUI_C1C		0x100	// Cursor1 Control Reg.
#define	ITE8181_GUI_C1O		0x101	// Cursor1 Offset Reg.
#define	ITE8181_GUI_C1F		0x102	// Cursor1 Feature Reg.
#define	ITE8181_GUI_C1SAH	0x103	// Icon Map Address MSB
#define	ITE8181_GUI_C1SAL	0x108	// Icon Map Address LSB
#define	ITE8181_GUI_C1LPX	0x109	// Cursor Clipping X Coord Reg.
#define	ITE8181_GUI_C1LPY	0x10a	// Cursor Clipping Y Coord Reg.
#define	ITE8181_GUI_CC0R0	0x110	// Cursor Color 0 Reg.
#define	ITE8181_GUI_CC0R1	0x111
#define	ITE8181_GUI_CC0R2	0x112
#define	ITE8181_GUI_CC0R3	0x113
#define	ITE8181_GUI_CC1R0	0x114	// Cursor Color 1 Reg.
#define	ITE8181_GUI_CC1R1	0x115
#define	ITE8181_GUI_CC1R2	0x116
#define	ITE8181_GUI_CC1R3	0x117
#define	ITE8181_GUI_CC2R0	0x118	// Cursor Color 2 Reg.
#define	ITE8181_GUI_CC2R1	0x119
#define	ITE8181_GUI_CC2R2	0x11a
#define	ITE8181_GUI_CC2R3	0x11b
#define ITE8181_GUI_C1XC0	0x120	// cursor 1 X coord bits[7:0]
#define ITE8181_GUI_C1XC1	0x121	// cursor 1 X coord bits[11:8]
#define ITE8181_GUI_C1YC0	0x122	// cursor 1 Y coord bits[7:0]
#define ITE8181_GUI_C1YC1	0x123	// cursor 1 Y coord bits[11:8]

/* ITE8181 Extension Mode-A Reg. (starting from Graphic Base addr) */
#define ITE8181_GRAPHIC8(index)	(&ite8181_graphic_base[(index)])
#define ITE8181_EMA_READ(index, var) \
	{\
		*ITE8181_GRAPHIC8(ITE8181_EMA_EXAX) = (index); \
		var = *ITE8181_GRAPHIC8(ITE8181_EMA_EXADATA); \
	}
#define ITE8181_EMA_WRITE(index, val) \
	{\
		*ITE8181_GRAPHIC8(ITE8181_EMA_EXAX) = (index); \
		*ITE8181_GRAPHIC8(ITE8181_EMA_EXADATA) = (val); \
	}

#define ITE8181_EMA_EXAX	0x03d6	// Extension Controller Index Reg.
#define ITE8181_EMA_EXADATA	0x03d7	// Extension Controller Data.

#define ITE8181_EMA_ENABLEEMA	0x0b	// Extension Index Enable Reg.
#define 	ITE8181_EMA_ENABLEPASS	0xec
#define 	ITE8181_EMA_DISABLEPASS	0xce

/* ITE8181 LCD Controller Timming Reg. */
#define ITE8181_EMA_HSIZE	0x80	// LCD Controler H size Reg.
#define ITE8181_EMA_HALIGN	0x81	// LCD H Align Adjust Reg.
#define ITE8181_EMA_HRETRACE	0x82	// LCD H Retrace Adjust Reg.
#define ITE8181_EMA_HADJUST	0x83	// LCD H Adjust Reg.
#define ITE8181_EMA_HSYNCDELAY	0x84	// LCD HSYNC Delay Reg.
#define ITE8181_EMA_VSIZE	0x85	// LCD V size Reg.
#define ITE8181_EMA_VSYNC_DELAY	0x86	// LCD VSYNC Delay Reg.
#define ITE8181_EMA_OVERFLOW	0x87	// LCD Overflow Reg.
#define ITE8181_EMA_MODULATION	0x88	// LCD Modulation Reg.
#define ITE8181_EMA_EXTMODE	0x89	// LCD Ext Mode Tuning Reg.
#define ITE8181_EMA_VALIGNA	0x8a	// LCD V Align Adjust Reg A(350)
#define ITE8181_EMA_VALIGNB	0x8b	// LCD V Align Adjust Reg B(400)
#define ITE8181_EMA_VALIGNC	0x8c	// LCD V Align Adjust Reg C(>=480)
#define ITE8181_EMA_VRETRACE	0x8d	// LCD V Retrace adjusr Reg.
#define ITE8181_EMA_VOVERFLOW	0x8e	// LCD V Adjust Overflow Reg.

/* ITE8181 LCD Controller Reg. */
#define ITE8181_EMA_TYPE	0x90	// LCD Type Select Reg.
#define ITE8181_EMA_CONTROL	0x91	// LCD Controller Reg.
#define ITE8181_EMA_PINSEL	0x92	// LCD Controller Pin Select Reg.
#define ITE8181_EMA_MISCCTL	0x93	// LCD Misc Control Reg.

/* ITE8181 LCD COntroller Power Management Register */
#define ITE8181_EMA_LCDPOWER	0x98
#define 	ITE8181_LCDSTANDBY	0x20	// LCD S/W Standby
#define ITE8181_EMA_LCDPOWERSEQ	0x9a
#define		ITE8181_PUP2		0x80	// Panel Power UP phase 2
#define		ITE8181_PUP1		0x40	// Panel Power UP phase 1
#define		ITE8181_PUP0		0x20	// Panel Power UP phase 0
#define		ITE8181_PDP2		0x10	// Panel Power DOWN phase 2
#define		ITE8181_PDP1		0x08	// Panel Power DOWN phase 1
#define		ITE8181_PDP0		0x04	// Panel Power DOWN phase 0
#define ITE8181_EMA_LCDPOWERSTAT 0x9b
#define		ITE8181_PPTOBEOFF	0x0	// Panel Power to be OFF
#define		ITE8181_PPTOBEON	0x1	// Panel Power to be ON

volatile unsigned int * ite8181_config_base;
volatile unsigned char *ite8181_memory_base;
volatile unsigned char *ite8181_gui_base;
volatile unsigned char *ite8181_graphic_base;

static int probed = 0;
static int backlight = -1;
static int contrast = -1;
static int lcdpower = -1;
static int standby_reg_standby, standby_reg_init, standby_reg;

int get_ite8181_backlight(void)
{
	return backlight;
}

int ite8181_backlight(int n)
{
	if (!probed)
		return 1;
	if (backlight != n) {
		if(n == 0)
			;
			// Turn the backlight off.
		else
			;
			// Turn the backlight on.
	}
	backlight = n;
	return 0;
}

int get_ite8181_contrast(void)
{
	return contrast;
}

int ite8181_contrast(int n)
{
	if (!probed)
		return 1;
	if (contrast != n) {
		;// Turn the lcd on and some contrast.
	}
	contrast = n;
	return 0;
}

int get_ite8181_lcdpower(void)
{
	return lcdpower;
}

void ite8181_wakeup(void)
{
#if 0
	int sreg = standby_reg_standby;
	int lcds;
	int lcdp;

	*ITE8181_CONF(ITE8181_STANDBY) = sreg;
	sreg &= ~ITE8181_DATA_CLOCKSTOP;
	*ITE8181_CONF(ITE8181_STANDBY) = sreg;
	//udelay(25000);
	sreg &= ~(ITE8181_DATA_GATEPLL2IN|ITE8181_DATA_GATEPLL1IN);
	*ITE8181_CONF(ITE8181_STANDBY) = sreg;
	//udelay(25000);

	ITE8181_EMA_WRITE(ITE8181_EMA_ENABLEEMA, ITE8181_EMA_ENABLEPASS);
	ITE8181_EMA_READ(ITE8181_EMA_LCDPOWER, lcdp);
	lcdp &= ~ITE8181_LCDSTANDBY;
	ITE8181_EMA_WRITE(ITE8181_EMA_LCDPOWER, lcdp);

	ITE8181_EMA_READ(ITE8181_EMA_LCDPOWERSTAT, lcds);
	lcds |= ITE8181_PPTOBEON;
	ITE8181_EMA_WRITE(ITE8181_EMA_LCDPOWERSTAT, lcds);
	//udelay(800000);
	ITE8181_EMA_WRITE(ITE8181_EMA_ENABLEEMA, ITE8181_EMA_DISABLEPASS);

	sreg |= (standby_reg|ITE8181_DATA_DACPOWERON);
	*ITE8181_CONF(ITE8181_STANDBY) = sreg;

	sreg &= ~(ITE8181_DATA_PALETTESTBY|ITE8181_DATA_PLL2_PWDOWN);
	*ITE8181_CONF(ITE8181_STANDBY) = sreg;

	sreg &= ~(ITE8181_DATA_CURSORSTBY|ITE8181_DATA_LINESTBY
		  |ITE8181_DATA_BITBLTSTBY);
	*ITE8181_CONF(ITE8181_STANDBY) = sreg;
#endif
}

void ite8181_standby(void)
{
#if 0
	int sreg = *ITE8181_CONF(ITE8181_STANDBY);
	int lcds;
	int lcdp;

	ITE8181_EMA_WRITE(ITE8181_EMA_ENABLEEMA, ITE8181_EMA_ENABLEPASS);
	ITE8181_EMA_READ(ITE8181_EMA_LCDPOWER, lcdp);
	lcdp |= ITE8181_LCDSTANDBY;
	ITE8181_EMA_WRITE(ITE8181_EMA_LCDPOWER, lcdp);

	ITE8181_EMA_READ(ITE8181_EMA_LCDPOWERSTAT, lcds);
	lcds &= ~ITE8181_PPTOBEON;
	ITE8181_EMA_WRITE(ITE8181_EMA_LCDPOWERSTAT, lcds);
	//udelay(800000);

	ITE8181_EMA_WRITE(ITE8181_EMA_ENABLEEMA, ITE8181_EMA_DISABLEPASS);

	sreg |= ITE8181_DATA_PALETTESTBY|ITE8181_DATA_PLL2_PWDOWN;
	*ITE8181_CONF(ITE8181_STANDBY) = sreg;

	sreg  |= ITE8181_DATA_DACCLKSTOP;
	sreg  &= ~ITE8181_DATA_DACPOWERON;
	*ITE8181_CONF(ITE8181_STANDBY) = sreg;

	sreg |= (ITE8181_DATA_CURSORSTBY|ITE8181_DATA_LINESTBY
		  |ITE8181_DATA_BITBLTSTBY);
	*ITE8181_CONF(ITE8181_STANDBY) = sreg;

	sreg |= ITE8181_DATA_CLOCKSTOP;
	*ITE8181_CONF(ITE8181_STANDBY) = sreg;
#endif
}

int ite8181_lcdpower(int on)
{
	int lcdp;
	int lcds;
	int lcdseq;

	if (!probed)
		return 1;
	if (lcdpower != on) {

		ITE8181_EMA_WRITE(ITE8181_EMA_ENABLEEMA, ITE8181_EMA_ENABLEPASS);
		ITE8181_EMA_READ(ITE8181_EMA_LCDPOWER, lcdp);
		ITE8181_EMA_READ(ITE8181_EMA_LCDPOWERSTAT, lcds);
		ITE8181_EMA_READ(ITE8181_EMA_LCDPOWERSEQ, lcdseq);
#ifdef DEBUG
		printk(KERN_INFO "ite8181_lcdpower(%d) LCDPOWER(0) = %x\n", on, lcdp);
		printk(KERN_INFO "ite8181_lcdpower(%d) LCDPOWERSTAT(0) = %x\n", on, lcds);
		printk(KERN_INFO "ite8181_lcdpower(%d) LCDPOWERSEQ(0) = %x\n", on, lcdseq);
#endif /* DEBUG */
		if (on) {
			// lcd power on
			lcdp &= ~ITE8181_LCDSTANDBY;
			ITE8181_EMA_WRITE(ITE8181_EMA_LCDPOWER, lcdp);
			lcds |= ITE8181_PPTOBEON;
			ITE8181_EMA_WRITE(ITE8181_EMA_LCDPOWERSTAT, lcds);
			//udelay(800000);
#ifdef DEBUG
			ITE8181_EMA_READ(ITE8181_EMA_LCDPOWER, lcdp);
			ITE8181_EMA_READ(ITE8181_EMA_LCDPOWERSTAT, lcds);
			printk(KERN_INFO "ITE8181_EMA_LCDPOWER(1) = %x\n", lcdp);
			printk(KERN_INFO "ITE8181_EMA_LCDPOWERSTAT(1) = %x\n", lcds);
#endif /* DEBUG */
		} else {
			// lcd power off
			lcdp |= ITE8181_LCDSTANDBY;
			ITE8181_EMA_WRITE(ITE8181_EMA_LCDPOWER, lcdp);
			lcds &= ~ITE8181_PPTOBEON;
			ITE8181_EMA_WRITE(ITE8181_EMA_LCDPOWERSTAT, lcds);
			//udelay(800000);
#ifdef DEBUG
			ITE8181_EMA_READ(ITE8181_EMA_LCDPOWER, lcdp);
			ITE8181_EMA_READ(ITE8181_EMA_LCDPOWERSTAT, lcds);
			printk(KERN_INFO "ITE8181_EMA_LCDPOWER(1) = %x\n", lcdp);
			printk(KERN_INFO "ITE8181_EMA_LCDPOWERSTAT(1) = %x\n", lcds);
#endif /* DEBUG */
		}
		ITE8181_EMA_WRITE(ITE8181_EMA_ENABLEEMA, ITE8181_EMA_DISABLEPASS);
	}
	lcdpower = on;
	return 0;
}

void ite8181_setup(void)
{
	// setup base address
	ite8181_config_base =
	       (volatile unsigned int *)(VIDEORAM_BASE + ((8192 -1) * 1024));
	ite8181_memory_base =
	       (volatile unsigned char *)
		       (VIDEORAM_BASE + *ITE8181_CONF(ITE8181_MBA));
	ite8181_gui_base =
	       (volatile unsigned char *)
		       (VIDEORAM_BASE + *ITE8181_CONF(ITE8181_GBA));
	ite8181_graphic_base =
	       (volatile unsigned char *)
		       (VIDEORAM_BASE + *ITE8181_CONF(ITE8181_SBA));
#ifdef DEBUG
	printk(KERN_INFO "ite8181_config_base = %08x\n",
		(unsigned int)ite8181_config_base);
	printk(KERN_INFO "ite8181_memory_base = %08x\n",
		(unsigned int)ite8181_memory_base);
	printk(KERN_INFO "ite8181_gui_base = %08x\n",
		(unsigned int)ite8181_gui_base);
	printk(KERN_INFO "ite8181_graphic_base = %08x\n",
		(unsigned int)ite8181_graphic_base);
#endif
	// probe
	if (*ITE8181_CONF(ITE8181_ID) != ITE8181_DATA_ID) {
#ifndef DEBUG
		probed = 0;
		return;
#else /* DEBUG */
		printk(KERN_INFO "ite8181_setup: ID (%08x) != %08x\n",
			*ITE8181_CONF(ITE8181_ID),
			ITE8181_DATA_ID);
#endif /* DEBUG */
	}
	probed = 1;
	printk(KERN_INFO "ite8181: attached\n");

	// save some register status
	standby_reg_init = standby_reg_standby = 
	standby_reg = *ITE8181_CONF(ITE8181_STANDBY);

	// Disable Hardware Cursor (drawn by WinCE -- kei */
	*ITE8181_GUI8(ITE8181_GUI_C1C) = 0; // Cursor 1 Control Reg.
#ifdef DEBUG
	printk(KERN_INFO "ITE8181_GUI8(ITE8181_GUI_C1C) = %08x\n",
	    ITE8181_GUI8(ITE8181_GUI_C1C));
	printk(KERN_INFO "*ITE8181_CONF(ITE8181_STANDBY) = %08x\n",
	    *ITE8181_CONF(ITE8181_STANDBY));
#endif
	ite8181_lcdpower(1);
	ite8181_contrast(7);
	ite8181_backlight(1);
}
