/*
 * linux/arch/mips/vr41xx/time.c
 *
 * VR41XX timer interrupt using real-time clock.
 *
 * Copyright (C) 1991, 1992, 1995  Linus Torvalds
 * Copyright (C) 1999 Bradley D. LaRonde
 * Copyright (C) 2000 Michael Klar
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/pm.h>
#include <asm/vr41xx.h>

extern int add_irq_action(int irq, struct irqaction *new);
extern volatile unsigned long wall_jiffies;
extern rwlock_t xtime_lock;

#define USECS_PER_JIFFY (1000000/HZ)
#define COUNTS_PER_JIFFY ((32768 + HZ/2) / HZ)

spinlock_t rtc_lock = SPIN_LOCK_UNLOCKED;

#ifdef CONFIG_RTC
unsigned long epoch_adj;
#else
#define epoch_adj 0
#endif

// per VR41xx docs, bad data can be read if between 2 counts
static inline unsigned short read_time_reg(volatile unsigned short *reg)
{
	unsigned short value;
	do {
		value = *reg;
		barrier();
	} while (value != *reg);
	return value;
}

unsigned long get_rtc_time(volatile unsigned short *reg)
{
	unsigned short regh, regm, regl;

	// why this crazy order, you ask?  to guarantee that neither m
	// nor l wrap before all 3 read
	do {
		regm = read_time_reg(reg + 1);
		barrier();
		regh = read_time_reg(reg + 2);
		barrier();
		regl = read_time_reg(reg);
	} while (regm != read_time_reg(reg + 1));
	return ((regh << 17) | (regm << 1) | (regl >> 15)) + epoch_adj;
}

void set_rtc_time(unsigned long settime, volatile unsigned short *reg)
{
	unsigned short intreg;
	unsigned long flags, timeval = settime - epoch_adj;

	spin_lock_irqsave(&rtc_lock, flags);
	intreg = *VR41XX_RTCINTREG & 0x05;
	barrier();
	*reg = timeval << 15;
	*(reg + 1) = timeval >> 1;
	*(reg + 2) = timeval >> 17;
	barrier();
	// assume that any ints that just triggered are invalid, since the
	// time value is written non-atomically in 3 separate regs
	*VR41XX_RTCINTREG = 0x05 ^ intreg;
	spin_unlock_irqrestore(&rtc_lock, flags);
}

// must be called with ints disabled:
//
static unsigned long do_gettimeoffset(void)
{
	unsigned short count;
	unsigned long offset;

	count = read_time_reg(VR41XX_RTCL1CNTLREG);
	if (count == 1)
		offset = 0;
	else
		offset = (COUNTS_PER_JIFFY - count + 1) * 1000000 / 32768;
	// detect if the counter wrapped, but int not serviced yet, being
	// careful not to adjust if int happen after count was read just now
	if (*VR41XX_RTCINTREG & 0x0002 && (jiffies == wall_jiffies) && count != 2)
		offset += USECS_PER_JIFFY;
	return offset;
}

// This version of gettimeofday has about 30us resolution
void do_gettimeofday(struct timeval *tv)
{
	unsigned long flags, lost;

	read_lock_irqsave(&xtime_lock, flags);
	*tv = xtime;
	tv->tv_usec += do_gettimeoffset();

	// xtime is atomically updated in timer_bh, wall_jiffies is
	// updated in timer_bh, jiffies is updated in timer int, so this
	// will be nonzero if the timer bottom half hasn't executed yet:
	lost = jiffies - wall_jiffies;

	read_unlock_irqrestore(&xtime_lock, flags);

	while (lost--)
		tv->tv_usec += USECS_PER_JIFFY;

	while (tv->tv_usec >= 1000000) {
		tv->tv_usec -= 1000000;
		tv->tv_sec++;
	}
}

void do_settimeofday(struct timeval *tv)
{
	write_lock_irq(&xtime_lock);
	// undo whatever correction gettimeofday would have done
	tv->tv_usec -= do_gettimeoffset();
	tv->tv_usec -= (jiffies - wall_jiffies) * USECS_PER_JIFFY;

	if (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		tv->tv_sec--;
	}

	xtime = *tv;
	time_adjust = 0;		// stop active adjtime()
	time_status |= STA_UNSYNC;
	time_maxerror = NTP_PHASE_LIMIT;
	time_esterror = NTP_PHASE_LIMIT;
	write_unlock_irq(&xtime_lock);
}

static unsigned long last_rtc_update;

/*
 * timer_interrupt() needs to keep up the real-time clock,
 * as well as call the "do_timer()" routine every clocktick
 */
static void timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	// Clear the interrupt.
	*VR41XX_RTCINTREG = 0x2;

	do_timer(regs);

	// If we have an externally synchronized Linux clock, then update
	// CMOS clock accordingly every ~11 minutes.
	if ((time_status & STA_UNSYNC) == 0 &&
	    xtime.tv_sec > last_rtc_update + 660) {
		set_rtc_time(xtime.tv_sec, VR41XX_ETIMELREG);
		last_rtc_update = xtime.tv_sec;
	}
}

struct irqaction timer_action = {
	timer_interrupt, SA_INTERRUPT, 0,
	"timer", NULL, NULL
};

#ifdef CONFIG_PM
static int pm_time_request(struct pm_dev *dev, pm_request_t rqst, void *data)
{
	unsigned long flags;

	switch (rqst) {
	case PM_SUSPEND:
		disable_irq(VR41XX_IRQ_INT1);
		break;
	case PM_RESUME:
		write_lock_irqsave(&xtime_lock, flags);
		enable_irq(VR41XX_IRQ_INT1);
		xtime.tv_sec = get_rtc_time(VR41XX_ETIMELREG);
		xtime.tv_usec = 0;
		// this shouldn't be necessary, but just to be safe...
		*VR41XX_RTCL1LREG = COUNTS_PER_JIFFY;
		*VR41XX_RTCL1HREG = 0;
		write_unlock_irqrestore(&xtime_lock, flags);
		break;
	}
	return 0;
}

static int __init pm_time_init(void)
{
	pm_register(PM_SYS_DEV, PM_SYS_UNKNOWN, pm_time_request);
	return 0;
}

__initcall(pm_time_init);
#endif

void __init time_init(void)
{
	// Set default time near beginning of Linux VR epoch.
	xtime.tv_sec = get_rtc_time(VR41XX_ETIMELREG);
	xtime.tv_usec = 0;

	// Use RTCLong1 for the system timer.
	// It has it's own cpu interrupt, is not T-Clock dependent,
	// and has sufficient resolution.

	// Set the RTCLong1 counter (32.768kHz) to expire in 1 / HZ second
	*VR41XX_RTCL1LREG = COUNTS_PER_JIFFY;
	*VR41XX_RTCL1HREG = 0;

	// and ack any pending ints
	*VR41XX_RTCINTREG = 0x2;

	// Grab the IRQ for the cpu interrupt, not the ICU interrupt
	// The RTCLong1 ICU interrupt is always left unmasked
	// This can't use request_irq() because it's too early for kmalloc
	add_irq_action(VR41XX_IRQ_INT1, &timer_action);

}
