/*
 * linux/arch/mips/vr41xx/vrc4173-ps2mouse.c
 *
 * Platform support for NEC Harrier.
 *
 * Copyright (C) 2000 Michael R. McDonald
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * The PS2 code originally came from drivers/char/kc_kbd.c
 *
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/pc_keyb.h>
#include <linux/kbd_kern.h>
#include <linux/kbd_ll.h>
#include <linux/random.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <asm/io.h>
#include <asm/keyboard.h>
#include <asm/vr41xx.h>
#include <asm/vrc4173.h>

#include <asm/vr41xx-platdep.h>

extern struct pci_dev *vrc4173_dev;

extern spinlock_t vrc4173_controller_lock;

void vrc4173_turn_on_device(struct pci_dev *pdev, int device);
void vrc4173_turn_off_device(struct pci_dev *pdev, int device);
int vrc4173_PS2_wait_for_input(struct pci_dev *pdev, int channel);
void vrc4173_PS2_write_data(struct pci_dev *pdev, int channel, unsigned short data);

/*
 *	PS/2 Auxiliary Device
 */

#define AUX_RECONNECT 170 /* scancode when ps2 device is plugged (back) in */
 
static struct aux_queue *queue;	/* Mouse data buffer. */
static int aux_count = 0;
/* used when we send commands to the mouse that expect an ACK. */
static unsigned char mouse_reply_expected = 0;

#define AUX_INTS_OFF (KBD_MODE_KCC | KBD_MODE_DISABLE_MOUSE | KBD_MODE_SYS | KBD_MODE_KBD_INT)
#define AUX_INTS_ON  (KBD_MODE_KCC | KBD_MODE_SYS | KBD_MODE_MOUSE_INT | KBD_MODE_KBD_INT)

#define MAX_RETRIES	60		/* some aux operations take long time*/

char *vrc4173_PS2_init_mouse(struct pci_dev *pdev, int channel)
{
    long ioaddr;

    ioaddr = pci_resource_start (pdev, 0);

    /* enable the port and turn off interrupts */
    outw(0x0, ioaddr + VRC4173_PS2CH1CTRL + channel);
    
    /*
     * Enable the mouse
     */
    vrc4173_PS2_write_data(pdev, channel, KBD_CCMD_MOUSE_ENABLE);
#ifdef INITIALIZE_MOUSE
    vrc4173_PS2_write_data(pdev, channel, AUX_SET_SAMPLE);
    if (vrc4173_PS2_wait_for_input(pdev, channel) != KBD_REPLY_ACK)
	return "Set sample rate: no ACK";

    vrc4173_PS2_write_data(pdev, channel, 100);
    if (vrc4173_PS2_wait_for_input(pdev, channel) != KBD_REPLY_ACK)
	return "Set sample rate = 100: no ACK";

    vrc4173_PS2_write_data(pdev, channel, AUX_SET_RES);
    if (vrc4173_PS2_wait_for_input(pdev, channel) != KBD_REPLY_ACK)
	return "Set resolution: no ACK";

    vrc4173_PS2_write_data(pdev, channel, 3);
    if (vrc4173_PS2_wait_for_input(pdev, channel) != KBD_REPLY_ACK)
	return "Set resolution = 3: no ACK";

    vrc4173_PS2_write_data(pdev, channel, AUX_SET_SCALE21);
    if (vrc4173_PS2_wait_for_input(pdev, channel) != KBD_REPLY_ACK)
	return "Set scale: no ACK";
#endif
    vrc4173_PS2_write_data(pdev, channel, AUX_ENABLE_DEV);
    if (vrc4173_PS2_wait_for_input(pdev, channel) != KBD_REPLY_ACK)
	return "Enable device: no ACK";

    vrc4173_PS2_write_data(pdev, channel, AUX_INTS_ON);
    vrc4173_PS2_wait_for_input(pdev, channel);

    outw(0x8, ioaddr + VRC4173_PS2CH1CTRL + channel);
    return NULL; /* success */
}

static inline void handle_mouse_event(unsigned char scancode)
{
	if (mouse_reply_expected) {
		if (scancode == AUX_ACK) {
			mouse_reply_expected--;
			return;
		}
		mouse_reply_expected = 0;
	}
	else if(scancode == AUX_RECONNECT){
		queue->head = queue->tail = 0;  /* Flush input queue */
		vrc4173_PS2_init_mouse(vrc4173_dev, 0x20);  /* ping the mouse :) */
		return;
	}

	add_mouse_randomness(scancode);
	if (aux_count) {
		int head = queue->head;

		queue->buf[head] = scancode;
		head = (head + 1) & (AUX_BUF_SIZE-1);
		if (head != queue->tail) {
			queue->head = head;
			kill_fasync(&queue->fasync, SIGIO, POLL_IN);
			wake_up_interruptible(&queue->proc_list);
		}
	}
}

void vrc4173_handle_ps2_mouse(struct pci_dev *pdev, int index)
{
    long ioaddr;
    unsigned short status, data;
    int channel = 0x20;

    ioaddr = pci_resource_start (pdev, 0);

    while ((status = inw(ioaddr + VRC4173_PS2CH1CTRL + channel)) & 0x1) {
	data = inw(ioaddr + VRC4173_PS2CH1DATA + channel);
	NEC_HARRIER_DISPLAY_LEDS(data);
	handle_mouse_event(data);
	printk("mouse 0x%04x\n", data);
    }
}

/*
 * Put bytes from input queue to buffer.
 */

static unsigned char get_from_queue(void)
{
	unsigned char result;
	unsigned long flags;

	spin_lock_irqsave(&vrc4173_controller_lock, flags);
	result = queue->buf[queue->tail];
	queue->tail = (queue->tail + 1) & (AUX_BUF_SIZE-1);
	spin_unlock_irqrestore(&vrc4173_controller_lock, flags);
	return result;
}

static inline int queue_empty(void)
{
	return queue->head == queue->tail;
}

/*
 * Send a byte to the mouse.
 */
static void aux_write_dev(int val)
{
	unsigned long flags;

	spin_lock_irqsave(&vrc4173_controller_lock, flags);
	vrc4173_PS2_write_data(vrc4173_dev, 0x20, KBD_CCMD_WRITE_MOUSE);
	vrc4173_PS2_write_data(vrc4173_dev, 0x20, val);
	spin_unlock_irqrestore(&vrc4173_controller_lock, flags);
}

static int fasync_aux(int fd, struct file *filp, int on)
{
	int retval;

	retval = fasync_helper(fd, filp, on, &queue->fasync);
	if (retval < 0)
		return retval;
	return 0;
}


/*
 * Random magic cookie for the aux device
 */
#define AUX_DEV ((void *)queue)

static int release_aux(struct inode * inode, struct file * file)
{
	fasync_aux(-1, file, 0);
	if (--aux_count)
		return 0;
	vrc4173_PS2_write_data(vrc4173_dev, 0x20, AUX_INTS_OFF);    /* Disable controller ints */
	vrc4173_PS2_write_data(vrc4173_dev, 0x20, KBD_CCMD_MOUSE_DISABLE);
	vrc4173_turn_off_device(vrc4173_dev, PS2AUX_DEV);
	return 0;
}

/*
 * Install interrupt handler.
 * Enable auxiliary device.
 */

static int open_aux(struct inode * inode, struct file * file)
{
	if (aux_count++) {
		return 0;
	}
	vrc4173_turn_on_device(vrc4173_dev, PS2AUX_DEV);
	queue->head = queue->tail = 0;		/* Flush input queue */
	vrc4173_PS2_init_mouse(vrc4173_dev, 0x20);

	return 0;
}

static ssize_t read_aux(struct file * file, char * buffer,
			size_t count, loff_t *ppos)
{
	DECLARE_WAITQUEUE(wait, current);
	ssize_t i = count;
	unsigned char c;

	if (queue_empty()) {
		if (file->f_flags & O_NONBLOCK)
			return -EAGAIN;
		add_wait_queue(&queue->proc_list, &wait);
repeat:
		set_current_state(TASK_INTERRUPTIBLE);
		if (queue_empty() && !signal_pending(current)) {
			schedule();
			goto repeat;
		}
		current->state = TASK_RUNNING;
		remove_wait_queue(&queue->proc_list, &wait);
	}
	while (i > 0 && !queue_empty()) {
		c = get_from_queue();
		put_user(c, buffer++);
		i--;
	}
	if (count-i) {
		file->f_dentry->d_inode->i_atime = CURRENT_TIME;
		return count-i;
	}
	if (signal_pending(current))
		return -ERESTARTSYS;
	return 0;
}

/*
 * Write to the aux device.
 */

static ssize_t write_aux(struct file * file, const char * buffer,
			 size_t count, loff_t *ppos)
{
	ssize_t retval = 0;

	if (count) {
		ssize_t written = 0;

		if (count > 32)
			count = 32; /* Limit to 32 bytes. */
		do {
			char c;
			get_user(c, buffer++);
			aux_write_dev(c);
			written++;
		} while (--count);
		retval = -EIO;
		if (written) {
			retval = written;
			file->f_dentry->d_inode->i_mtime = CURRENT_TIME;
		}
	}

	return retval;
}

/* No kernel lock held - fine */
static unsigned int aux_poll(struct file *file, poll_table * wait)
{
	poll_wait(file, &queue->proc_list, wait);
	if (!queue_empty())
		return POLLIN | POLLRDNORM;
	return 0;
}

struct file_operations psaux_fops = {
	read:		read_aux,
	write:		write_aux,
	poll:		aux_poll,
	open:		open_aux,
	release:	release_aux,
	fasync:		fasync_aux,
};

/*
 * Initialize driver.
 */
static struct miscdevice psaux_mouse = {
	PSMOUSE_MINOR, "psaux", &psaux_fops
};

int __init vrc4173_psaux_init(void)
{
	misc_register(&psaux_mouse);
	queue = (struct aux_queue *) kmalloc(sizeof(*queue), GFP_KERNEL);
	memset(queue, 0, sizeof(*queue));
	queue->head = queue->tail = 0;
	init_waitqueue_head(&queue->proc_list);

	return 0;
}
