/*
 * linux/arch/mips/vr41xx/vrc4173-tpanel.c
 *
 * Platform support for NEC Harrier.
 *
 * Copyright (C) 2000 Michael R. McDonald
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * The touch panel code originally came from arch/mips/vr41xx/tpanel.c
 *
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/tpanel.h>
#include <asm/io.h>
#include <asm/vr41xx.h>
#include <asm/vrc4173.h>

#define DEBUG_TIMING_PARAMS

#ifdef DEBUG_TIMING_PARAMS
#include <linux/proc_fs.h>
#endif

#include <asm/vr41xx-platdep.h>

extern struct pci_dev *vrc4173_dev;

void vrc4173_turn_on_device(struct pci_dev *pdev, int device);
void vrc4173_turn_off_device(struct pci_dev *pdev, int device);

// BUFFSIZE can be increased, but it must be a power of 2
#define BUFFSIZE 128

#define MAX_READ_SIZE 0xFFFF

struct tpanel_status {
    unsigned int	buffer[BUFFSIZE];
    unsigned char	head;
    unsigned char	tail;
    unsigned char	lostflag;
    unsigned char	lastcontact;
    wait_queue_head_t 	wait;
    struct fasync_struct *fasyncptr;
    int			active;
#ifdef DEBUG_TIMING_PARAMS
    unsigned short	orig_scanint;
    unsigned short	orig_settletime;
    unsigned int	dropped_count;
    unsigned int	overflow_count;
    unsigned int	incompletescan_count;
    unsigned int	dualpagevalid_count;
    unsigned int	unhandledint_count;
#endif
};

static struct tpanel_status tpanel;

static void tpanel_add_data(unsigned int data1, unsigned int data2)
{
    unsigned int errcnt;
/* printk("d1 %d d2 %d\n", data1, data2); */
    tpanel.buffer[tpanel.head++] = data1 | ((unsigned int)tpanel.lostflag << 28);
    tpanel.buffer[tpanel.head++] = data2;
    tpanel.head &= (BUFFSIZE - 1);
    if (tpanel.head == tpanel.tail) {
	errcnt = tpanel.buffer[tpanel.tail] & 0xff;
	if (!(tpanel.buffer[tpanel.tail] & 0x80000000)) {
	    errcnt++;
	    if (!(errcnt & 0xff))
		errcnt = 0xff;
	}
	errcnt |= tpanel.buffer[tpanel.tail] & 0x10000000;
	tpanel.tail += 2;
	tpanel.tail &= (BUFFSIZE - 1);
	tpanel.buffer[tpanel.tail] = (tpanel.buffer[tpanel.tail] & 0xffffff00) |
	    0x20000000 | errcnt;
#ifdef DEBUG_TIMING_PARAMS
	tpanel.dropped_count++;
#endif
    }
    tpanel.lostflag = 0;
}

static void vrc4173_and_or_reg(struct pci_dev *pdev, int reg, unsigned short and, unsigned short or)
{
    long ioaddr;
    unsigned short bits;

    ioaddr = pci_resource_start (pdev, 0);
    cli();
    bits = inw(ioaddr + reg);
    bits &= and;
    bits |= or;
    outw(bits, ioaddr + reg);
    sti();
}

void vrc4173_handle_tpanel(struct pci_dev *pdev, int index)
{
    long ioaddr;
    unsigned int cschange = 0, data1 = 0, data2 = 0;
    unsigned short intreg, pendown;

    ioaddr = pci_resource_start (pdev, 0);

    while ((intreg = (inw(ioaddr + VRC4173_PIUINTREG) & 0x807d) | cschange)) {
	pendown = inw(ioaddr + VRC4173_PIUCNTREG) & 0x2000;
	if (tpanel.lastcontact != (pendown >> 13)) {
	    cschange = 1;
	}
	tpanel.lastcontact = pendown >> 13;
	if (intreg & VRC4173_PIUINTREG_PENCHG) {
	    outw(VRC4173_PIUINTREG_PENCHG, ioaddr + VRC4173_PIUINTREG);
	    vrc4173_and_or_reg(pdev, VRC4173_PIUINTREG, 0xffff, VRC4173_PIUINTREG_PENCHG);
	}

	// PADDLOST interrupt: datapoint lost, just report it
	// (after processing data and pen-down)
	if (((intreg & (VRC4173_PIUINTREG_PADDLOST | VRC4173_PIUINTREG_PADPAGE0 |
			VRC4173_PIUINTREG_PADPAGE1)) == VRC4173_PIUINTREG_PADDLOST) &&
	    !(cschange && pendown)) {
	    outw(VRC4173_PIUINTREG_PADDLOST, ioaddr + VRC4173_PIUINTREG);
	    barrier();
	    outw(0x0022, ioaddr + VRC4173_PIUCNTREG);
	    barrier();

	    pendown = inw(ioaddr + VRC4173_PIUCNTREG) & 0x2000;
	    if (tpanel.lastcontact != (pendown >> 13))
		cschange = 1;
	    tpanel.lastcontact = pendown >> 13;

	    if (pendown) {
		outw(0x0062, ioaddr + VRC4173_PIUCNTREG);
		barrier();
		outw(0x0066, ioaddr + VRC4173_PIUCNTREG);	// force manual scan
	    } else {
		outw(0x0322, ioaddr + VRC4173_PIUCNTREG);
		barrier();
		outw(0x0326, ioaddr + VRC4173_PIUCNTREG);	// resume autoscan
	    }

	    tpanel.lostflag = 1;
#ifdef DEBUG_TIMING_PARAMS
	    if ((intreg & (VRC4173_PIUINTREG_PADPAGE1 | VRC4173_PIUINTREG_PADPAGE0))
		== (VRC4173_PIUINTREG_PADPAGE1 | VRC4173_PIUINTREG_PADPAGE0))
		tpanel.overflow_count++;
	    else
		tpanel.incompletescan_count++;
#endif
	}

	// Panel contact state change interrupt
	// If pen-up, do data ints first
	if (cschange && (pendown || !(intreg & (VRC4173_PIUINTREG_PADPAGE0 |
						VRC4173_PIUINTREG_PADPAGE1)))) {
	    data1 = (pendown | 0x4000) << 17;
	    tpanel_add_data(data1, 0);
	    cschange = 0;
	}

	// PADCMD and PADADP shouldn't happen
	if (intreg & (VRC4173_PIUINTREG_PADCMD | VRC4173_PIUINTREG_PADADP)) {
	    outw(VRC4173_PIUINTREG_PADCMD | VRC4173_PIUINTREG_PADADP,
		 ioaddr + VRC4173_PIUINTREG);
#ifdef DEBUG_TIMING_PARAMS
	    tpanel.unhandledint_count++;
#endif
	}

	// Data buffer interrupt
	if (intreg & VRC4173_PIUINTREG_PADPAGE0) {
#ifdef DEBUG_TIMING_PARAMS
	    if (intreg & VRC4173_PIUINTREG_PADPAGE1)
		tpanel.dualpagevalid_count++;
#endif
	    if ((intreg & (VRC4173_PIUINTREG_PADPAGE1 | VRC4173_PIUINTREG_OVP))
		!= (VRC4173_PIUINTREG_PADPAGE1 | VRC4173_PIUINTREG_OVP)) {
		data1 = 0x40000000 |
		    ((inw(ioaddr + VRC4173_PIUPB01REG) & MAX_READ_SIZE) << 18) |
		    ((inw(ioaddr + VRC4173_PIUPB00REG) & MAX_READ_SIZE) << 8);
		data2 = ( inw(ioaddr + VRC4173_PIUPB03REG) << 20) |
		    ((inw(ioaddr + VRC4173_PIUPB02REG) & MAX_READ_SIZE) << 10) |
		    ( inw(ioaddr + VRC4173_PIUPB04REG) & MAX_READ_SIZE);
		tpanel_add_data(data1, data2);

		outw(VRC4173_PIUINTREG_PADPAGE0, ioaddr + VRC4173_PIUINTREG);
	    }
	}
	if (intreg & VRC4173_PIUINTREG_PADPAGE1) {
	    data1 = 0x40000000 |
		((inw(ioaddr + VRC4173_PIUPB11REG) & MAX_READ_SIZE) << 18) |
		((inw(ioaddr + VRC4173_PIUPB10REG) & MAX_READ_SIZE) << 8);
	    data2 = ( inw(ioaddr + VRC4173_PIUPB13REG) << 20) |
		((inw(ioaddr + VRC4173_PIUPB12REG) & MAX_READ_SIZE) << 10) |
		( inw(ioaddr + VRC4173_PIUPB14REG) & MAX_READ_SIZE);
	    tpanel_add_data(data1, data2);

	    outw(VRC4173_PIUINTREG_PADPAGE1, ioaddr + VRC4173_PIUINTREG);
	}
	break;
    }

    if (data1) {
	add_mouse_randomness(data1);
	if (data2)
	    add_mouse_randomness(data2);
	wake_up_interruptible(&tpanel.wait);

	if (tpanel.fasyncptr)
	    kill_fasync(&tpanel.fasyncptr, SIGIO, POLL_IN);
    }
    outw(0x0326, ioaddr + VRC4173_PIUCNTREG);	// sometimes bits get reset in this reg
    barrier();					// so we just reinitialize it
}

static int fasync_tpanel(int fd, struct file *filp, int on)
{
    int retval;

    retval = fasync_helper(fd, filp, on, &tpanel.fasyncptr);
    if (retval < 0)
	return retval;
    return 0;
}

static int close_tpanel(struct inode * inode, struct file * file)
{
    long ioaddr = pci_resource_start(vrc4173_dev, 0);

    fasync_tpanel(-1, file, 0);
    if (--tpanel.active)
	return 0;
    // set for standby
    outw(0, ioaddr + VRC4173_MPIUINTREG);
    outw(0, ioaddr + VRC4173_PIUCNTREG);
    vrc4173_turn_off_device(vrc4173_dev, PIU_DEV);

    return 0;
}

static int open_tpanel(struct inode * inode, struct file * file)
{
    long ioaddr = pci_resource_start(vrc4173_dev, 0);

    if (tpanel.active++)
	return 0;
    vrc4173_and_or_reg(vrc4173_dev, VRC4173_SELECTREG, ~VRC4173_SELECT_GPIO_H, 0);
    outw(0, ioaddr + VRC4173_MPIUINTREG);
    vrc4173_turn_on_device(vrc4173_dev, PIU_DEV);

    while (((inw(ioaddr + VRC4173_PIUCNTREG) >> 10) & 0x7) != VRC4173_PIUCNTREG_STATE_DISABLE)
	;
    
    // power up outputs and reset the PIU
    vrc4173_and_or_reg(vrc4173_dev, VRC4173_PIUCNTREG, 0xffff, VRC4173_PIUCNTREG_PIUPWR);

    while (((inw(ioaddr + VRC4173_PIUCNTREG) >> 10) & 0x7) != VRC4173_PIUCNTREG_STATE_STANDBY)
	;
    
    tpanel.tail = tpanel.head;
    tpanel.lastcontact = 0;

    outw(0x007d, ioaddr + VRC4173_PIUINTREG);	// clear any pending ints

    outw(333, ioaddr + VRC4173_PIUSIVLREG);	// set interval to .01 sec default
    outw(25, ioaddr + VRC4173_PIUSTBLREG);
    outw(1<<12, ioaddr + VRC4173_PIUCMDREG);
    barrier();

    outw(0x007d, ioaddr + VRC4173_MPIUINTREG);	// unmask interrupts
    barrier();

    outw(0x0326, ioaddr + VRC4173_PIUCNTREG);	// autoscan, sequence enabled
    return 0;
}

/*
 *  Read touch panel data.
 *
 *  Data format:
 *  unsigned short status: bit 15 = xyz data valid (0 means contact state only)
 *                         bit 14 = pen contact state (1 means contact)
 *                         bit 13 = soft data lost flag: if this is 1, this data
 *                                  is valid (if bit 15 is 1), but data was lost
 *                                  between this point and the previous point
 *                                  due to a buffer overrun in the driver
 *                         bit 12 = hard data lost flag: if this is 1, this point
 *                                  is valid (if bit 15 is 1), but data was lost
 *                                  between this point and the previous point
 *                                  due to hardware error
 *                         bits 11-8 = reserved
 *                         bits 7-0 = count of how many data packet lost, if hard
 *                                    or soft error is falgged
 *  unsigned short x+ raw data (if status:15 = 1)
 *  unsigned short x- raw data (if status:15 = 1)
 *  unsigned short y+ raw data (if status:15 = 1)
 *  unsigned short y- raw data (if status:15 = 1)
 *  unsigned short z (pressure) raw data (if status:15 = 1)
 *
 *  x+, x-, y+, and y- are limited to range 0-1023 for this hardware.  Each +/-
 *  pair is somewhat redundant: the + value can be used as is, but using the
 *  difference (eg. (x+) - (x-)) will produce a more accurate result.  Further
 *  manipulation and/or statistical analysis may be required for best accuracy.
 *
 *  No calibration is done on the driver side, that is expected to be done on
 *  the user side.
 */

#define MAX_READ_SIZE 0x6FFF

static ssize_t read_tpanel(struct file * file,
			   char * buffer, size_t count, loff_t *ppos)
{
    unsigned int data1, data2;
    size_t retcnt = 0;

    if (count < 12)
	return -EINVAL;

    // We need to access the circular buffer atomic to
    // anything else that will read or write it
    cli();

    while (tpanel.head == tpanel.tail) {
	if (file->f_flags & O_NONBLOCK) {
	    sti();
	    return -EAGAIN;
	}
	interruptible_sleep_on(&tpanel.wait);
	// sleep_on will sti for us
	if (signal_pending(current))
	    return -ERESTARTSYS;
	cli();
    }
    do {
	data1 = tpanel.buffer[tpanel.tail++];
	data2 = tpanel.buffer[tpanel.tail++];
	tpanel.tail &= BUFFSIZE - 1;

	sti();

	if (!access_ok(VERIFY_WRITE, buffer, 12))
	    return -EFAULT;

	__put_user((((data1 >> 16) & 0xf000) | (data1 & 0x00ff)) ^ 0x8000,
		   (short*)(buffer + retcnt));
	__put_user((data1 >> 18) & MAX_READ_SIZE, (short*)(buffer + retcnt + 2));
	__put_user((data1 >> 8) & MAX_READ_SIZE, (short*)(buffer + retcnt + 4));
	__put_user((data2 >> 20) & MAX_READ_SIZE, (short*)(buffer + retcnt + 6));
	__put_user((data2 >> 10) & MAX_READ_SIZE, (short*)(buffer + retcnt + 8));
	__put_user(data2 & MAX_READ_SIZE, (short*)(buffer + retcnt + 10));

	retcnt += 12;
	if (retcnt > count - 12)
	    return retcnt;

	cli();
    } while (tpanel.head != tpanel.tail);

    sti();

    return retcnt;
}

static unsigned int poll_tpanel(struct file *file, poll_table * wait)
{
    int comp;

    poll_wait(file, &tpanel.wait, wait);
    cli();
    comp = (tpanel.head != tpanel.tail);
    sti();
    if (comp)
	return POLLIN | POLLRDNORM;
    return 0;
}

static int ioctl_tpanel(struct inode * inode, struct file * file,
			unsigned int cmd, unsigned long arg)
{
    struct scanparam *sparm;
    unsigned int i1, i2;
    long ioaddr = pci_resource_start(vrc4173_dev, 0);

    sparm = (struct scanparam *)arg;

    switch (cmd) {
    case TPGETSCANPARM:
	i1 = inw(ioaddr + VRC4173_PIUSIVLREG) * 30;	// PIUSIVLREG is in 30us units
	if (put_user(i1, &sparm->interval))
	    return -EFAULT;
	i2 = inw(ioaddr + VRC4173_PIUSTBLREG) * 30;	// PIUSTBLREG is in 30us units
	if (put_user(i2, &sparm->settletime))
	    return -EFAULT;
	return 0;
    case TPSETSCANPARM:
	if (get_user(i1, &sparm->interval))
	    return -EFAULT;
	i1 = (i1 + 15) / 30;
	if (i1 > 0x07ff)
	    return -EINVAL;
	if (get_user(i2, &sparm->settletime))
	    return -EFAULT;
	i2 = (i2 + 15) / 30;
	if (i2 > 0x003f)
	    return -EINVAL;
	outw((unsigned short) i1, ioaddr + VRC4173_PIUSIVLREG);
	outw((unsigned short) i2, ioaddr + VRC4173_PIUSTBLREG);
	__put_user(i1 * 30, &sparm->interval);
	__put_user(i2 * 30, &sparm->settletime);
	return 0;
    default:
	return -EINVAL;
    }
}

struct file_operations vrc4173_tpanel_fops = {
    read:	read_tpanel,
    poll:	poll_tpanel,
    ioctl:	ioctl_tpanel,
    open:	open_tpanel,
    release:	close_tpanel,
    fasync:	fasync_tpanel,
};

static struct miscdevice vrc4173_tpanel = {
    VR41XX_TPANEL_MINOR, "vrc4173tpanel", &vrc4173_tpanel_fops
};

#ifdef DEBUG_TIMING_PARAMS
static int get_piuinfo(char *page, char **start, off_t off,
		       int count, int *eof, void *data)
{
    int len;
    long ioaddr = pci_resource_start(vrc4173_dev, 0);

    // return lots of nifty diagnostic info
    // buffer is PAGE_SIZE long, we're OK for all page sizes VRC4173 supports
    len = sprintf(page,
		  "piucntreg:            0x%04hx\n"
		  "piusivlreg:           0x%04hx\n"
		  "piustblreg:           0x%04hx\n"
		  "orig piusivlreg:      0x%04hx\n"
		  "orig piustblreg:      0x%04hx\n"
		  "dropped data count:   %d\n"
		  "page overflow count:  %d\n"
		  "bad scan count:       %d\n"
		  "dual page count:      %d\n"
		  "unhandled int count:  %d\n",
		  inw(ioaddr + VRC4173_PIUCNTREG),
		  inw(ioaddr + VRC4173_PIUSIVLREG),
		  inw(ioaddr + VRC4173_PIUSTBLREG),
		  tpanel.orig_scanint, tpanel.orig_settletime, tpanel.dropped_count,
		  tpanel.overflow_count, tpanel.incompletescan_count,
		  tpanel.dualpagevalid_count, tpanel.unhandledint_count);

    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len > count) len = count;
    if (len < 0) len = 0;
    return len;
}
#endif

int __init vrc4173_tpanel_init(struct pci_dev *pdev, int channel)
{
    long ioaddr;

    ioaddr = pci_resource_start (pdev, 0);

    vrc4173_turn_on_device(pdev, PIU_DEV);
    barrier();
#ifdef DEBUG_TIMING_PARAMS

    tpanel.orig_scanint = inw(ioaddr + VRC4173_PIUSIVLREG);
    tpanel.orig_settletime = inw(ioaddr + VRC4173_PIUSTBLREG);
    tpanel.dropped_count = 0;
    tpanel.overflow_count = 0;
    tpanel.incompletescan_count = 0;
    tpanel.dualpagevalid_count = 0;
    tpanel.unhandledint_count = 0;

    create_proc_read_entry("piuinfo", 0, NULL, get_piuinfo, NULL);
#endif

    tpanel.active = 0;
    tpanel.head = tpanel.tail = 0;
    init_waitqueue_head(&tpanel.wait);
    tpanel.fasyncptr = NULL;
    printk("VRC4173 touch panel initialized.\n");
    misc_register(&vrc4173_tpanel);

    // clocks remained masked until the touch panel device is opened
    vrc4173_turn_off_device(pdev, PIU_DEV);

    return 0;
}
