/*
 * linux/drivers/video/vr4181.c
 *
 * NEC VR4181 LCD Controller
 * 
 * Copyright (C) 1999 Bradley D. LaRonde.
 *
 * Based on sed1354 frame buffer by Bradley D. LaRonde.
 *
 * I've only done 8bpp for now (for example the hardware palette management),
 * but I've left the code for the other bpps in for when I (or someone else)
 * eventually does them.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file COPYING in the main directory of this archive for
 * more details.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/pm.h>
#include <video/fbcon.h>
#include <video/fbcon-mfb.h>
#include <video/fbcon-cfb2.h>
#include <video/fbcon-cfb4.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-mac.h>
#include <asm/addrspace.h>
#include <asm/vr41xx.h>
#include <asm/uaccess.h>

// #define VR4181FB_DEBUG KERN_INFO

#define DEVICE_NAME "vr4181fb"

#define VR4181FB_XRES_MAX     320
#define VR4181FB_YRES_MAX     320
#define VR4181FB_BPP_MAX      8
#define VR4181FB_MEM_LEN_MAX  ((VR4181FB_XRES_MAX * VR4181FB_YRES_MAX * VR4181FB_BPP_MAX) / 8)

// VR4181FB_MEM_START should probably should be in video config options, since it depends on
// location of kernel data.
#define VR4181FB_MEM_START    (KSEG1 + 0x1000)

static struct fb_var_screeninfo var_def = {
	0, 0, 0, 0,
	0, 0, 0, 0,
	{0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0},
	0, FB_ACTIVATE_NOW, -1, -1, 0, 20000, 64, 64, 32, 32, 64, 2,
	0, FB_VMODE_NONINTERLACED, {0,0,0,0,0,0}
};

static int inverse_def;
static int currcon;
static struct display disp;
static struct fb_info my_fb_info;

#ifdef FBCON_HAS_CFB8
// fbcon uses 16 palette indicies
// support mapping through dispsw_data
static union {
	unsigned char cfb8[16];
} fbcon_cmap;
#endif

// helper functions

static u_long get_line_length(int xres_virtual, int bpp)
{
	return (xres_virtual * bpp + 7) >> 3;
}

static void get_fix_screeninfo(struct fb_fix_screeninfo *fix, struct fb_var_screeninfo *var)
{
	// Return fixed display parameters.
	memset(fix, 0, sizeof(struct fb_fix_screeninfo));
	strcpy(fix->id, DEVICE_NAME);
	fix->smem_start = VR4181FB_MEM_START;
	fix->smem_len = VR4181FB_MEM_LEN_MAX;
	fix->type = FB_TYPE_PACKED_PIXELS;
	fix->type_aux = 0;
	switch (var->bits_per_pixel) {
		case 1:
			fix->visual = FB_VISUAL_MONO01;
			break;

		default:
			fix->visual = FB_VISUAL_PSEUDOCOLOR;
			break;
	}
	fix->ywrapstep = 1;
	fix->xpanstep = 1;
	fix->ypanstep = 1;
	fix->line_length = get_line_length(var->xres_virtual, var->bits_per_pixel);
}

static void get_color_bitfields(struct fb_var_screeninfo *var)
{
	var->red.offset =    0;
	var->red.length =    var->bits_per_pixel;
	var->green.offset =  0;
	var->green.length =  var->bits_per_pixel;
	var->blue.offset =   0;
	var->blue.length =   var->bits_per_pixel;
	var->transp.offset = 0;
	var->transp.length = 0;

	var->red.msb_right =    0;
	var->green.msb_right =  0;
	var->blue.msb_right =   0;
	var->transp.msb_right = 0;
}

static int my_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
                         u_int *transp, struct fb_info *info)
{
	// Read a single color register and split it into
	// colors/transparent. Return != 0 for invalid regno.
	unsigned int val;

	if (regno > 255)
		return 1;

	// Get palette entry from hardware.
	*VR41XX_CPINDCTREG = regno;
	val = *VR41XX_CPALDATREG;
	
	// Break out palette entry into individual 16-bit values.
	*red =   (val & 0x3f000) >> 2;
	*green = (val & 0xfc0) << 4;
	*blue =  (val & 0x3f) << 10;

	if(info->disp->inverse) {
		*red = 0xfc00 - *red;
		*green = 0xfc00 - *green;
		*blue = 0xfc00 - *blue;
	}

#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": gcr %d %x/%x/%x\n",
		regno, *red, *green, *blue);
#endif
	return 0;
}

static int my_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
                         u_int transp, struct fb_info *info)
{
	// Set a single color register. The values supplied are already
	// rounded down to the hardware's capabilities (according to the
	// entries in the var structure). Return != 0 for invalid regno.

	// So it says anyway.
	// From what I can tell, drivers/video/fbcmap.c passes in
	// un-rounded-down 16-bit values.
	
#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": scr %d %x/%x/%x\n",
		regno, red, green, blue);
#endif

	if (regno > 255)
		return 1;

#ifdef FBCON_HAS_CFB8
	if( regno < 16 ) {
		// maintain index to rgb map for fbcon (accessed through dispsw_data)
		fbcon_cmap.cfb8[regno] =
			((red & 0xe000) >> 8)
			| ((green & 0xe000) >> 11)
			| ((blue & 0xc000) >> 14);
	}
#endif

	if(info->disp->inverse) {
		// Invert the intesities.
		red = 0xffff - red;
		green = 0xffff - green;
		blue = 0xffff - blue;
	}

	// Set the hardware palette register.
	*VR41XX_CPINDCTREG = regno;
	*VR41XX_CPALDATREG =
		((red & 0xfc00) << 2)
		| ((green & 0xfc00) >> 4)
		| ((blue & 0xfc00) >> 10);

	return 0;
}

static void do_install_cmap(int con, struct fb_info *info)
{
#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": do_install_cmap con %d currcon %d info %p conmaplen %d\n",
		con, currcon, info, fb_display[con].cmap.len);
#endif

	if (con != currcon) {
#ifdef VR4181FB_DEBUG
		printk(VR4181FB_DEBUG DEVICE_NAME ": Not installing any cmap: con != currcon\n");
#endif
		return;
	}

	// Is there a cmap installed on the target console?
	if (fb_display[con].cmap.len) {
#ifdef VR4181FB_DEBUG
		printk(VR4181FB_DEBUG DEVICE_NAME ": Installing the custom cmap.\n");
#endif
		// Install the custom cmap.
		fb_set_cmap(&fb_display[con].cmap, 1, my_setcolreg, info);
	}
	else {
#ifdef VR4181FB_DEBUG
		printk(VR4181FB_DEBUG DEVICE_NAME ": Installing the default cmap.\n");
#endif
		// Install the default cmap.
		fb_set_cmap(fb_default_cmap(1 << fb_display[con].var.bits_per_pixel), 1,
			my_setcolreg, info);
	}
}

static void vr4181fb_set_hardware(struct fb_var_screeninfo *var)
{
	int fblen;
	unsigned short m_signal, pre_scalar, color_depth, panel_color, panel_bus;

#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": set_hardware %dx%d-%d@%d\n",
		var->xres, var->yres, var->bits_per_pixel, 0);
#endif

	// Set up LCD mode.
	*VR41XX_LCDGPMODE = 0x0;

	// Set fifo, turn clock off, and set clock polarities.
	*VR41XX_LCDCTRLREG = 0xe0;

	// No interrupts, clear interrupts.
	*VR41XX_LCDIMSKREG = 0x0;
	*VR41XX_LCDINRQREG = 0x0;

	// Set horizontal total and visible.
	*VR41XX_HRTOTALREG = (var->xres + 10) / 2;
	*VR41XX_HRVISIBREG = var->xres / 8;

	// Set load clock (LOCLK).
	*VR41XX_LDCLKSTREG = (var->xres - 2) / 2; 
	*VR41XX_LDCLKNDREG = (var->xres + 4) / 2;

	// Set vertical total and visible.
	*VR41XX_VRTOTALREG = var->yres; // maybe need a little extra here?
	*VR41XX_VRVISIBREG = var->yres;

	// Set frame clock (FLM).
	*VR41XX_FHSTARTREG = 0;
	*VR41XX_FVSTARTREG = 0;
	*VR41XX_FHENDREG =   (var->xres + 8) / 2;
	*VR41XX_FVENDREG =   0;

	// Set M signal, gclck pre-scalar, color depth, and panel type.
	m_signal = 0 << 8;
	pre_scalar = 0 << 4;
	color_depth = ((var->bits_per_pixel == 8) ? 3 : var->bits_per_pixel / 2) << 2;
	panel_color = ((var->grayscale || (var->bits_per_pixel == 1)) ? 1 : 0) << 1;
	panel_bus = (var->bits_per_pixel == 8) ? 1 : 0;
	*VR41XX_LCDCFGREG0 = m_signal | pre_scalar | color_depth | panel_color | panel_bus;

	// Set gclk to Hpck ratio.
	*VR41XX_LCDCFGREG1 = var->pixclock;

	// Set frame buffer start address.
	*VR41XX_FBSTAD1REG = VR4181FB_MEM_START & 0xffff;
	*VR41XX_FBSTAD2REG = (VR4181FB_MEM_START & 0xffff0000) >> 16;

	// Set frame buffer end address.
	fblen = var->yres * get_line_length(var->xres_virtual, var->bits_per_pixel);
	*VR41XX_FBNDAD1REG = (VR4181FB_MEM_START + fblen) & 0xffff;
	*VR41XX_FBNDAD2REG = ((VR4181FB_MEM_START + fblen) & 0xffff0000) >> 16;

	// Turn LCD clock on.
	*VR41XX_LCDCTRLREG = 0xe8;

	// Turn LCD power on.
	display_power(1);

#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG "LCD Controller and related registers:\n");
	printk(VR4181FB_DEBUG
		"HT   %.4x  HV   %.4x  LS   %.4x  LE   %.4x  VT   %.4x  VV   %.4x  FVS  %.4x  FVE  %.4x\n",
		*VR41XX_HRTOTALREG, *VR41XX_HRVISIBREG, *VR41XX_LDCLKSTREG, *VR41XX_LDCLKNDREG,
		*VR41XX_VRTOTALREG, *VR41XX_VRVISIBREG, *VR41XX_FVSTARTREG, *VR41XX_FVENDREG);
	printk(VR4181FB_DEBUG
		"LCTR %.4x  LIRQ %.4x  LCG0 %.4x  LCG1 %.4x  FBS1 %.4x  FBS2 %.4x  FBE1 %.4x  FBE2 %.4x\n",
 		*VR41XX_LCDCTRLREG, *VR41XX_LCDINRQREG, *VR41XX_LCDCFGREG0, *VR41XX_LCDCFGREG1,
		*VR41XX_FBSTAD1REG, *VR41XX_FBSTAD2REG, *VR41XX_FBNDAD1REG, *VR41XX_FBNDAD2REG);
	printk(VR4181FB_DEBUG
		"FHS  %.4x  FHE  %.4x  PCG1 %.4x  PCG2 %.4x  PC0S %.4x  PC0E %.4x  PC0H %.4x  PCSM %.4x\n",
		*VR41XX_FHSTARTREG, *VR41XX_FHENDREG,   *VR41XX_PWRCONREG1, *VR41XX_PWRCONREG2,
		*VR41XX_PCS0STRA,   *VR41XX_PCS0STPA,   *VR41XX_PCS0HIA,    *VR41XX_PCSMODE);
	printk(VR4181FB_DEBUG
		"LGPM %.4x  GPM1 %.4x\n",
                *VR41XX_LCDGPMODE,  *VR41XX_GPMD1REG);
#endif
}


// framebuffer functions

static int my_fb_open(struct fb_info *info, int user)
{
	MOD_INC_USE_COUNT;
	return(0);
}
        
static int my_fb_release(struct fb_info *info, int user)
{
	MOD_DEC_USE_COUNT;
	return(0);
}

static int my_fb_get_fix(struct fb_fix_screeninfo *fix, int con,
	struct fb_info *info)
{
	struct fb_var_screeninfo *var;

	// No current console?
	if (con == -1)
		// Use default var info.
		var = &var_def;
	else
		// Use current console var info.
		var = &fb_display[con].var;

	// Fill in the fixed display params based on var.
	get_fix_screeninfo(fix, var);

	return 0;
}

static int my_fb_get_var(struct fb_var_screeninfo *var, int con,
	struct fb_info *info)
{
	// No current console?
	if (con == -1)
		// Use default var info.
		*var = var_def;
	else
		// Use current console var info.
		*var = fb_display[con].var;

	// Fill in the color bitfield info.
	get_color_bitfields(var);

	return 0;
}

static int my_fb_set_var(struct fb_var_screeninfo *var, int con,
	struct fb_info *info)
{
	int err, activate = var->activate;
	int oldxres, oldyres, oldvxres, oldvyres, oldbpp;
	u_long line_length;
	struct display *display;

#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": set_var con %d %dx%d-%d@%d\n",
		con, var->xres, var->yres, var->bits_per_pixel, 0);
#endif

	if (con >= 0)
		display = &fb_display[con];
	else
		display = &disp; // used during initialization

	// FB_VMODE_CONUPDATE and FB_VMODE_SMOOTH_XPAN are equal!
	// as FB_VMODE_SMOOTH_XPAN is only used internally

	if (var->vmode & FB_VMODE_CONUPDATE) {
		var->vmode |= FB_VMODE_YWRAP;
		var->xoffset = display->var.xoffset;
		var->yoffset = display->var.yoffset;
	}

	// Check/adjust the configuration.
	if (var->xres == 0 || var->yres == 0) {
		printk(KERN_ERR DEVICE_NAME
			": Error: unsupported resolution %dx%d.\n",
			var->xres, var->yres);
		return -EINVAL;
	}

	if (var->xres > var->xres_virtual)
		var->xres_virtual = var->xres;
	if (var->yres > var->yres_virtual)
		var->yres_virtual = var->yres;

	if (var->bits_per_pixel <= 1)
		var->bits_per_pixel = 1;
	else if (var->bits_per_pixel <= 2)
		var->bits_per_pixel = 2;
	else if (var->bits_per_pixel <= 4)
		var->bits_per_pixel = 4;
	else if (var->bits_per_pixel <= 8)
		var->bits_per_pixel = 8;
	else {
		printk(KERN_ERR DEVICE_NAME
			": Error: unsupported color depth %d bpp.\n",
			var->bits_per_pixel);
		return -EINVAL;
	}

	// Memory limit
	line_length = get_line_length(var->xres_virtual, var->bits_per_pixel);
	if (line_length * var->yres_virtual > VR4181FB_MEM_LEN_MAX) {
		printk(KERN_ERR DEVICE_NAME
			": Error: frame buffer size %ld exceeds %d maximum.\n",
			line_length * var->yres_virtual, VR4181FB_MEM_LEN_MAX);
		return -ENOMEM;
	}

	get_color_bitfields(var);

	if ((activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW) {
#ifdef VR4181FB_DEBUG
		printk(VR4181FB_DEBUG DEVICE_NAME ": set_var FB_ACTIVATE_NOW\n");
#endif
		oldxres = display->var.xres;
		oldyres = display->var.yres;
		oldvxres = display->var.xres_virtual;
		oldvyres = display->var.yres_virtual;
		oldbpp = display->var.bits_per_pixel;
		display->var = *var;

		if (oldxres != var->xres || oldyres != var->yres ||
			oldvxres != var->xres_virtual || oldvyres != var->yres_virtual ||
			oldbpp != var->bits_per_pixel) {

			struct fb_fix_screeninfo fix;

#ifdef VR4181FB_DEBUG
			printk(VR4181FB_DEBUG DEVICE_NAME
				": set_var FB_ACTIVATE_NOW settings changed\n");
#endif

			get_fix_screeninfo(&fix, var);

			display->screen_base = (char*)VR4181FB_MEM_START;
			display->visual = fix.visual;
			display->type = fix.type;
			display->type_aux = fix.type_aux;
			display->ypanstep = fix.ypanstep;
			display->ywrapstep = fix.ywrapstep;
			display->line_length = fix.line_length;
			display->can_soft_blank = 0;
			display->inverse = inverse_def;
			display->scrollmode = SCROLL_YREDRAW;
#ifdef FBCON_HAS_MAC
			display->dispsw = &fbcon_mac; /* it can handle everything */
#else
			switch (var->bits_per_pixel) {
#ifdef FBCON_HAS_MFB
				case 1:
					display->dispsw = &fbcon_mfb;
					break;
#endif
#ifdef FBCON_HAS_CFB2
				case 2:
					display->dispsw = &fbcon_cfb2;
					break;
#endif
#ifdef FBCON_HAS_CFB4
				case 4:
					display->dispsw = &fbcon_cfb4;
					break;
#endif
#ifdef FBCON_HAS_CFB8
				case 8:
					display->dispsw = &fbcon_cfb8;
					display->dispsw_data = fbcon_cmap.cfb8;
					break;
#endif
				default:
					display->dispsw = &fbcon_dummy;
					break;
			}
#endif

			if (my_fb_info.changevar)
				(*my_fb_info.changevar)(con);
		}

		if (oldbpp != var->bits_per_pixel) {
			if ((err = fb_alloc_cmap(&display->cmap, 0, 0)))
				return err;

#ifdef VR4181FB_DEBUG
			printk(VR4181FB_DEBUG DEVICE_NAME ": re-installing cmap: bpp changed.\n");
#endif
			do_install_cmap(con, info);
		}
	}

	// Set up the hardware.
	vr4181fb_set_hardware(var);

	return 0;
}

static int my_fb_pan_display(struct fb_var_screeninfo *var, int con,
			   struct fb_info *info)
{
#if 1
	return -EINVAL;
#else
	// Pan or Wrap the Display
	// This call looks only at xoffset, yoffset and the FB_VMODE_YWRAP flag

	if (var->vmode & FB_VMODE_YWRAP) {
		if (var->yoffset < 0 ||
			var->yoffset >= fb_display[con].var.yres_virtual ||
			var->xoffset)
			return -EINVAL;
	}
	else {
		if (var->xoffset+fb_display[con].var.xres >
			fb_display[con].var.xres_virtual ||
			var->yoffset+fb_display[con].var.yres >
			fb_display[con].var.yres_virtual)
			return -EINVAL;
	}

	fb_display[con].var.xoffset = var->xoffset;
	fb_display[con].var.yoffset = var->yoffset;

	if (var->vmode & FB_VMODE_YWRAP)
		fb_display[con].var.vmode |= FB_VMODE_YWRAP;
	else
		fb_display[con].var.vmode &= ~FB_VMODE_YWRAP;

	return 0;
#endif
}

static int my_fb_get_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": get_cmap %p kspc %d con %d, currcon %d info %p concmaplen %d\n",
		cmap, kspc, con, currcon, info, fb_display[con].cmap.len);
#endif

	if (con == currcon)
		return fb_get_cmap(cmap, kspc, my_getcolreg, info);
	else if (fb_display[con].cmap.len) // non default colormap?
		fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
	else
		fb_copy_cmap(fb_default_cmap(1<<fb_display[con].var.bits_per_pixel), cmap, kspc ? 0 : 2);

	return 0;
}

static int my_fb_set_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
	int err;
	struct display* display = &fb_display[con];

#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": set_cmap %p kspc %d con %d currcon %d info %p concmaplen %d\n",
		cmap, kspc, con, currcon, info, display->cmap.len);
#endif

	// no colormap allocated?
	if (!display->cmap.len) {
		err = fb_alloc_cmap(&display->cmap, 1 << display->var.bits_per_pixel, 0);
		if (err)
			return err;
	}

	// current console?
	if (con == currcon)
		return fb_set_cmap(cmap, kspc, my_setcolreg, info);
	else
		fb_copy_cmap(cmap, &display->cmap, kspc ? 0 : 1);

	return 0;
}

static int my_fb_ioctl(struct inode *inode, struct file *file, u_int cmd,
	u_long arg, int con, struct fb_info *info)
{
	int value;

	switch(cmd) {
		case FBIOGET_BACKLIGHT:
			value = display_get_backlight();
			put_user(value, (int*)arg);
			return 0;

		case FBIOPUT_BACKLIGHT:
			display_set_backlight(arg);
			return 0;

		case FBIOGET_CONTRAST:
			value = display_get_contrast();
			put_user(value, (int*)arg);
			return 0;

		case FBIOPUT_CONTRAST:
			display_set_contrast(arg);
			return 0;

#ifdef VR4181FB_DEBUG
		// bdl:  Hack to mess with LCD timings.
		// Should be moved to set_var I think.
		case 0:
			display_power(0);
			// Turn LCD clock off.
			*VR41XX_LCDCTRLREG = 0xe0;

			*VR41XX_LCDCFGREG0 = (unsigned short)arg;

			// Turn LCD clock on.
			*VR41XX_LCDCTRLREG = 0xe8;
			display_power(1);
			return 0;

		case 1:
			display_power(0);
			// Turn LCD clock off.
			*VR41XX_LCDCTRLREG = 0xe0;

			*VR41XX_LCDCFGREG1 = (unsigned short)arg;

			// Turn LCD clock on.
			*VR41XX_LCDCTRLREG = 0xe8;
			display_power(1);
			return 0;
#endif
	}

	return -EINVAL;
}


// console support

static int my_fbcon_switch(int con, struct fb_info *info)
{
#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": fbcon_switch con %d currcon %d currconmaplen %d info %p\n",
		con, currcon, fb_display[currcon].cmap.len, info);
#endif

	// Does this console have a colormap set?
	if (fb_display[currcon].cmap.len)
		// Save the colormap.
		fb_get_cmap(&fb_display[currcon].cmap, 1, my_getcolreg, info);

	currcon = con;

#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": installing cmap: console switch.\n");
#endif
	do_install_cmap(con, info);
	return 0;
}

static int my_fbcon_updatevar(int con, struct fb_info *info)
{
#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": fbcon_updatevar con %d currcon %d currconmaplen %d info %p\n",
		con, currcon, fb_display[currcon].cmap.len, info);
#endif
	// Update the `var' structure (called by fbcon.c)
	return 0;
}

static void my_fbcon_blank(int blank, struct fb_info *info)
{
#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": fbcon_blank blank %d info %p\n",
		blank, info);
#endif
	if(blank)
		display_power(0);
	else
		display_power(1);
}

#ifndef MODULE
int __init vr4181fb_setup(char *options)
{
	// Process "video=DEVICE_NAME:" boot argument options.
	// Exported for fbmem.c for compiled-in frame buffer driver.

	char *this_opt;

#ifdef VR4181FB_DEBUG	
	printk(VR4181FB_DEBUG DEVICE_NAME ": setup\n");
#endif

	if (!options || !*options)
		return 0;

#ifdef VR4181FB_DEBUG	
	printk(VR4181FB_DEBUG DEVICE_NAME ": setup options = %s\n", options);
#endif

	my_fb_info.fontname[0] = '\0';

	for(this_opt=strtok(options,","); this_opt; this_opt=strtok(NULL,",")) {
		if (!*this_opt) continue;
		
#ifdef VR4181FB_DEBUG	
		printk(VR4181FB_DEBUG DEVICE_NAME ": setup option %s\n", this_opt);
#endif

		if (!strncmp(this_opt, "xres:", 5))
			var_def.xres = var_def.xres_virtual =
				 simple_strtoul(this_opt + 5, NULL, 0);
		else if (!strncmp(this_opt, "yres:", 5))
			var_def.yres = var_def.yres_virtual =
				 simple_strtoul(this_opt + 5, NULL, 0);
		else if (!strncmp(this_opt, "bpp:", 4))
			var_def.bits_per_pixel = simple_strtoul(this_opt + 4, NULL, 0);
		else if (!strncmp(this_opt, "font:", 5))
			strncpy(my_fb_info.fontname, this_opt + 5, sizeof(my_fb_info.fontname)-1);
		else if (!strcmp(this_opt, "gray"))
			var_def.grayscale = 1;
		else if (!strcmp(this_opt, "inv"))
			// No "inverse" in fb_var_screeninfo, so use separate global.
			inverse_def = 1;
		else if (!strncmp(this_opt, "hpck:", 5))
			// HACK: store hpck setting in pixclock.
			var_def.pixclock = simple_strtoul(this_opt + 5, NULL, 0);
		else
			printk(KERN_WARNING DEVICE_NAME ": unrecognized option %s\n", this_opt);
	}
	return 0;
}
#endif

#ifdef CONFIG_PM
static int vr4181fb_pm_request(struct pm_dev *dev, pm_request_t rqst, void *data)
{
	switch (rqst) {
	case PM_SUSPEND:
#ifdef VR4181FB_DEBUG
		printk(VR4181FB_DEBUG DEVICE_NAME ": powering off\n");
#endif
		display_power(0);
		break;

	case PM_RESUME:
#ifdef VR4181FB_DEBUG
		printk(VR4181FB_DEBUG DEVICE_NAME ": powering on\n");
#endif
		display_power(1);
		break;
	}
	return 0;
}

static int __init vr4181fb_pm_init(void)
{
#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": pm_init\n");
#endif
	pm_register(PM_SYS_DEV, PM_SYS_UNKNOWN, vr4181fb_pm_request);
	return 0;
}

__initcall(vr4181fb_pm_init);
#endif

static struct fb_ops my_fb_ops = {
	owner: THIS_MODULE,
	fb_open: my_fb_open,
	fb_release: my_fb_release,
	fb_get_fix: my_fb_get_fix,
	fb_get_var: my_fb_get_var,
	fb_set_var: my_fb_set_var,
	fb_get_cmap: my_fb_get_cmap,
	fb_set_cmap: my_fb_set_cmap,
	fb_pan_display: my_fb_pan_display,
	fb_ioctl: my_fb_ioctl,
	fb_mmap: 0,
	fb_rasterimg: 0
};

int __init vr4181fb_init(void)
{
	// Exported for fbmem.c.

	int result;

#ifdef VR4181FB_DEBUG
	printk(VR4181FB_DEBUG DEVICE_NAME ": initializing\n");
#endif


	// Set up my_fb_info.
	strcpy(my_fb_info.modename, DEVICE_NAME);
	my_fb_info.changevar = NULL;
	my_fb_info.node = -1;
	my_fb_info.fbops = &my_fb_ops;
	my_fb_info.disp = &disp;
	my_fb_info.switch_con = &my_fbcon_switch;
	my_fb_info.updatevar = &my_fbcon_updatevar;
	my_fb_info.blank = &my_fbcon_blank;
	my_fb_info.flags = FBINFO_FLAG_DEFAULT;
	result = my_fb_set_var(&var_def, -1, &my_fb_info);
	if(result != 0) {
		printk(KERN_ERR DEVICE_NAME
			": Error setting default settings: result %d.\n",
			result);
		return result;
	}

	// Register this framebuffer.	
	result = register_framebuffer(&my_fb_info);
	if (result != 0)
	{
		printk(KERN_ERR DEVICE_NAME
			": Error registering frame buffer: result %d.\n",
			 result);
		return -EINVAL;
	}

	printk(KERN_INFO "fb%d: %s frame buffer, %dx%d, %d bpp, %ldk mem at 0x%.8x - 0x%.8x\n",
		GET_FB_IDX(my_fb_info.node), my_fb_info.modename, 
		var_def.xres, var_def.yres, var_def.bits_per_pixel,
		(var_def.yres * get_line_length(var_def.xres_virtual, var_def.bits_per_pixel)) >> 10,
		((*VR41XX_FBSTAD2REG << 16) | *VR41XX_FBSTAD1REG),
		((*VR41XX_FBNDAD2REG << 16) | *VR41XX_FBNDAD1REG));

	return 0;
}


#ifdef MODULE
int init_module(void)
{
	return vr4181fb_init();
}

void cleanup_module(void)
{
	unregister_framebuffer(&my_fb_info);
}
#endif /* MODULE */

